"""
Example code for estimaing matrix of resource utilzations from relative species abundances using Common Components Analysis.
Requires numpy and scipy for Python (v3).

Charles K. Fisher (2015)

Cite: “Habitat fluctuations drive species covariation in the human microbiota” by Charles K. Fisher, Thierry Mora, and Aleksandra M. Walczak
"""

import os, sys
import numpy as np

# Function Definitions #
def normalize(x):
    """ Compute compositional vector from vector of counts using pseudocount of 0.5. """
    return (x+0.5)/np.sum(x+0.5)

def alr(v):
    """ Compute additive log-ratio transform of the compositional vector v, referenced to position 0. """
    return np.log(v[1:]) - np.log(v[0])

def random_orthogonal_matrix(n):
    """ Generate a random orthogonal matrix. """
    A = np.random.normal(size=(n,n))
    Q, _ = np.linalg.qr(A)
    return Q

def recover_matrix(V0):
    """ Solve sparse matrix recovery problem to get matrix of niche volumes. """
    V = V0
    V = np.insert(V,0,0,axis=0)
    
    z = []
    for j in range(len(V0)):
        tmp = np.unique(np.append(-V0[:,j],0.0))
        z.append(sorted([[np.abs(t) + sum(np.abs(V0[:,j] + t)),t] for t in tmp])[0][1]) # L1 method
    
    return V + z

def get_common_components(Y_cov, p, max_steps=5000, skip=100, tolerance = 0.001):
    """ Estimate common components using gradient descent. """
    
    # compute PCA of mean covariance matrix (used to initialize gradient descent)
    Psi0 = np.mean(Y_cov,axis=0)
    eigvals, eigvecs = np.linalg.eig(Psi0)
    V0 = eigvecs
    S0 = np.array([eigvals for i in p])
    W0 = np.linalg.inv(V0.T)
    
    inv = [np.dot(np.dot(W0, np.diag(1.0/S0[i])), W0.T) for i in range(len(p))]
    eigs = [np.linalg.eigvalsh(mat) for mat in inv]
    like0 = np.dot(p,[np.trace(np.dot(Y_cov[i],inv[i])) - np.sum(np.log(eigs[i])) for i in range(len(p))])
    
    dS = S0 - S0
    dW = W0 - W0
    stepS = 0.0001
    stepW = 0.0001
    momentum = 0.95
    
    descent = []
    for t in range(max_steps):
        if t%2 == 0:
            # make a gradient descent step in S
            dS = np.array([np.diag(np.dot(np.dot(W0.T,Y_cov[i]),W0)) - S0[i] for i in range(len(p))]) + momentum*dS
            S1 = np.array([1.0/(1.0/S0[i] - stepS*dS[i]) for i in range(len(p))])
            
            inv = [np.dot(np.dot(W0, np.diag(1.0/S1[i])), W0.T) for i in range(len(p))]
            eigs = [np.linalg.eigvalsh(mat) for mat in inv]
            
            if np.min(eigs) > 0:
                like1 = np.dot(p,[np.trace(np.dot(Y_cov[i],inv[i])) - np.sum(np.log(eigs[i])) for i in range(len(p))])
                if like1 < like0:
                    # accept step and increase step size
                    like0 = like1
                    S0 = S1;
                    stepS = 1.1 * stepS
                else:
                    # reject step and decrease step size
                    stepS = 0.5 * stepS
                    dS = S0 - S0
            else:
                # reject step and decrease step size
                stepS = 0.5 * stepS
                dS = S0 - S0
        else:
            # make a gradient descent step in W
            dW = momentum * dW
            for i in range(len(p)):
                dW += p[i]*(np.dot(np.dot(Y_cov[i],W0),np.diag(1.0/S0[i])) - np.linalg.inv(W0).T)
            W1 = W0 - stepW * dW
            
            # normalize columns of W
            W1 = np.dot(W1,np.diag(1.0/np.linalg.norm(W1,axis=0)))
            
            inv = [np.dot(np.dot(W1, np.diag(1.0/S0[i])), W1.T) for i in range(len(p))]
            eigs = [np.linalg.eigvalsh(mat) for mat in inv]
            
            if np.min(eigs) > 0:
                like1 = np.dot(p,[np.trace(np.dot(Y_cov[i],inv[i])) - np.sum(np.log(eigs[i])) for i in range(len(p))])
                if like1 < like0:
                    # accept step and increase step size
                    like0 = like1
                    W0 = W1;
                    stepW = 1.1 * stepW
                else:
                    # reject step and decrease step size
                    stepW = 0.5 * stepW
                    dW = W0 - W0
            else:
                # reject step and decrease step size
                stepW = 0.5 * stepw
                dW = W0 - W0
    
        # periodically record value of objective function and check covergence
        if t%skip == 0:
            print('Percent Complete: {}\r'.format(100.0*t/max_steps),end='\r')
            descent.append(like0)
            if len(descent) > 2 and np.abs(descent[-1] - descent[-2]) < tolerance:
                print('\nConvergence tolerance reached')
                break;

    print('Percent Complete: 100.0')
    return [descent, W0, S0]


# Main Program #
if __name__ == "__main__":
    
    # read in file with species counts
    input = [line.strip().split() for line in open('hmp_top_species_counts.dat','r')]
    labels = [row[1] for row in input[1:]]
    env = sorted(list(set([row[1] for row in input[1:]])))
    
    # compute compositions and ALR transforms from counts
    X = np.array([normalize( np.array([float(y) for y in row[2:]])) for row in input[1:]])
    Y = np.array([alr(x) for x in X])
    
    # compute PCA
    Y_cov_pca = np.cov(Y.T)
    eigvals0, eigvecs0 = np.linalg.eig(Y_cov_pca)
    V_pca = recover_matrix(eigvecs0)
    
    # reshape arrays based on environments (i.e. body sites)
    Y_shaped = [np.array([Y[i] for i in range(len(Y)) if input[i+1][1] == e]) for e in env]
    
    # compute means, covariance matrices, and weights for the environments
    Y_cov = [np.cov(y.T) for y in Y_shaped]
    p = np.array([float(len(y)) for y in Y_shaped] / np.sum([len(y) for y in Y_shaped]))
    
    # compute CCA on randomized covariance matrices 'max_iter' times
    max_iter = 20
    for iter in range(max_iter):
        print("iteration: %d" % iter)
        
        # randomize covariance matrices
        Y_cov_rand = []
        for mat in Y_cov:
            rot = random_orthogonal_matrix(len(mat))
            eigvals = np.linalg.eigvalsh(mat)
            Y_cov_rand.append(np.dot(rot,np.dot(np.diag(eigvals),rot.T)))
        
        # estimate common components using gradient descent
        [descent,W0,S0] = get_common_components(Y_cov_rand,p,max_steps = 500000, tolerance = 0.0001)

        # sort common components
        lambda_cca = np.dot(Y,W0)
        tmp = [np.array([lambda_cca[i] for i in range(len(labels)) if labels[i] == e]) for e in env]
        lambda_ave = np.array([np.mean(t,axis=0) for t in tmp])
        lambda_var = np.array([np.var(t,axis=0) for t in tmp])
        score = np.var(lambda_ave,axis=0)/np.mean(lambda_var,axis=0)
        order = np.argsort(score)[::-1]
        
        # recover V matrix and common components
        V_cca = recover_matrix(np.array([v[order] for v in np.linalg.inv(W0).T]))
        W_ordered = [w[order] for w in W0]
        lambda_cca = [v[order] for v in lambda_cca]
        
        # output results to tab delimited files

        # create directory for output files if it doesn't already exist
        directory = 'common_components_output_randomization'
        if not os.path.exists(directory):
            os.makedirs(directory)
        
        # write out the randomly generated covariance matrices
        output = open(directory + '/covar_' + str(iter) + '.dat','w')
        for mat in Y_cov_rand:
            for row in mat:
                for col in row:
                    output.write("%.6f\t" % col)
                output.write("\n")
            output.write("\n")

        # write out the objective function during gradient descent
        output = open(directory + '/objective_function_' + str(iter) + '.dat','w')
        for x in descent:
            output.write("%.6f\n" % x)
        output.close()

        # write out the diagonal elements of the sigma (S0) matrices
        output = open(directory + '/sigma_' + str(iter) + '.dat','w')
        for i in range(len(S0)):
            output.write(env[i] + "\t")
            tmp = S0[i][order]
            for x in tmp:
                output.write("%.6f\t" % x)
            output.write("\n")
        output.close()

        # write out the estimate for the V matrix
        output = open(directory + '/V_matrix_' + str(iter) + '.dat','w')
        for row in V_cca:
            for col in row:
                output.write("%.6f\t" % col)
            output.write("\n")
        output.close()

        # write out the estimate for the V matrix from pca
        output = open(directory + '/V_pca_matrix_' + str(iter) + '.dat','w')
        for row in V_pca:
            for col in row:
                output.write("%.6f\t" % col)
            output.write("\n")
        output.close()

        # write out the estimate for the W matrix
        output = open(directory + '/W_matrix_' + str(iter) + '.dat','w')
        for row in W_ordered:
            for col in row:
                output.write("%.6f\t" % col)
            output.write("\n")
        output.close()

        # write out the lambdas
        output = open(directory + '/lambda_' + str(iter) + '.dat','w')
        for i in range(len(lambda_cca)):
            output.write(labels[i] + "\t")
            for col in lambda_cca[i]:
                output.write("%.6f\t" % col)
            output.write("\n")
        output.close()

